import ast
from typing import List, Tuple, Optional

import mysql.connector
from mysql.connector import Error

from pathfinder.core.repo.mysql_config_utility import parse_mysql_config

class MysqlNGDRepository:

    @classmethod
    def from_config_string(cls, config_str: str) -> "MysqlNGDRepository":
        cfg = parse_mysql_config(config_str)

        return cls(
            host=cfg.host,
            user=cfg.user,
            password=cfg.password,
            database=cfg.database,
            port=cfg.port,
            ssl_required=cfg.ssl_required,
        )

    def __init__(
        self,
        host: str = "arax-databases-mysql.rtx.ai",
        user: str = "public_ro",
        password: Optional[str] = None,
        database: str = "curie_ngd_v1_0_kg2_10_2",
        port: int = 3306,
        ssl_required: bool = True,
        connect_timeout: int = 10,
    ):
        self._conn = self._connect(
            host=host,
            user=user,
            password=password,
            database=database,
            port=port,
            ssl_required=ssl_required,
            connect_timeout=connect_timeout,
        )

    def _connect(self, **kwargs):
        ssl_required = kwargs.pop("ssl_required", True)
        try:
            conn = mysql.connector.connect(
                **kwargs,
                ssl_disabled=not ssl_required,
                autocommit=True,
            )
            return conn
        except Error as e:
            raise RuntimeError(f"MySQL connection failed: {e}")

    def close(self):
        try:
            if self._conn and self._conn.is_connected():
                self._conn.close()
        except Exception:
            pass

    def _ensure_connection(self):
        if not self._conn or not self._conn.is_connected():
            raise RuntimeError("MySQL connection is not available.")

    def get_curie_ngd(self, curie: str) -> List:
        self._ensure_connection()
        try:
            with self._conn.cursor() as cursor:
                query = "SELECT ngd FROM curie_ngd WHERE curie = %s"
                cursor.execute(query, (curie,))
                row = cursor.fetchone()
        except Exception as e:
            raise Exception(f"{e}, mysql host: {self._conn.server_host if hasattr(self._conn,'server_host') else 'unknown'}")

        if row and row[0] is not None:
            return ast.literal_eval(row[0])

        return []

    def get_curies_pmid_length(self, curies: List[str], limit: int = -1) -> List[Tuple[str, int]]:
        self._ensure_connection()
        if not curies:
            return []

        placeholders = ",".join(["%s"] * len(curies))

        if limit != -1:
            query = f"""
                SELECT curie, pmid_length
                FROM curie_ngd
                WHERE curie IN ({placeholders})
                ORDER BY pmid_length DESC
                LIMIT %s
            """
            params = tuple(curies) + (limit,)
        else:
            query = f"""
                SELECT curie, pmid_length
                FROM curie_ngd
                WHERE curie IN ({placeholders})
                ORDER BY pmid_length DESC
            """
            params = tuple(curies)

        try:
            with self._conn.cursor() as cursor:
                cursor.execute(query, params)
                rows = cursor.fetchall()
                return rows
        except Exception as e:
            raise Exception(f"{e}, mysql query failed")



from pathfinder.core.repo.MLRepo import MLRepo
from pathfinder.core.repo.NGDRepository import NGDRepository
from pathfinder.core.repo.NGDSortedNeighborsRepo import NGDSortedNeighborsRepo
from pathfinder.core.repo.NodeDegreeRepo import NodeDegreeRepo
from pathfinder.core.repo.PloverDBRepo import PloverDBRepo
from pathfinder.core.repo.MysqlNGDRepository import MysqlNGDRepository
from pathfinder.core.repo.MysqlNodeDegreeRepo import MysqlNodeDegreeRepo


def get_ngd_repo(ngd_url):
    if ngd_url.startswith("sqlite:"):
        return NGDRepository(ngd_url.removeprefix("sqlite:"))
    elif ngd_url.startswith("mysql:"):
        return MysqlNGDRepository.from_config_string(ngd_url)
    else:
        raise ValueError(f"Unknown ngd_url '{ngd_url}'.")

def get_degree_repo(degree_url):
    if degree_url.startswith("sqlite:"):
        return NodeDegreeRepo(degree_url.removeprefix("sqlite:"))
    elif degree_url.startswith("mysql:"):
        return MysqlNodeDegreeRepo.from_config_string(degree_url)
    else:
        raise ValueError(f"Unknown ngd_url '{degree_url}'.")


def get_repo(repo_name, plover_url, ngd_url, degree_url):
    if repo_name == "NGDSortedNeighborsRepo":
        return NGDSortedNeighborsRepo(
            PloverDBRepo(plover_url=plover_url, degree_repo=NodeDegreeRepo(degree_url)),
            get_degree_repo(degree_url),
            get_ngd_repo(ngd_url)
        )
    elif repo_name == "MLRepo":
        return MLRepo(
            PloverDBRepo(plover_url=plover_url, degree_repo=NodeDegreeRepo(degree_url)),
            get_degree_repo(degree_url),
            get_ngd_repo(ngd_url)
        )
    else:
        raise ValueError(f"Unknown animal_type '{repo_name}'.")



import json
from typing import Dict, List, Set, Optional

import mysql.connector
from mysql.connector import Error
from pathfinder.core.repo.mysql_config_utility import parse_mysql_config


class MysqlNodeDegreeRepo:

    @classmethod
    def from_config_string(cls, config_str: str) -> "MysqlNodeDegreeRepo":
        cfg = parse_mysql_config(config_str)

        return cls(
            host=cfg.host,
            user=cfg.user,
            password=cfg.password,
            database=cfg.database,
            port=cfg.port,
            ssl_required=cfg.ssl_required,
        )

    def __init__(
        self,
        host: str = "arax-databases-mysql.rtx.ai",
        user: str = "public_ro",
        password: Optional[str] = None,
        database: str = "curie_ngd_v1_0_kg2_10_2",
        port: int = 3306,
        ssl_required: bool = True,
        connect_timeout: int = 10,
    ):
        self._conn = self._connect(
            host=host,
            user=user,
            password=password,
            database=database,
            port=port,
            ssl_required=ssl_required,
            connect_timeout=connect_timeout,
        )

    def _connect(self, **kwargs):
        ssl_required = kwargs.pop("ssl_required", True)
        try:
            conn = mysql.connector.connect(
                **kwargs,
                ssl_disabled=not ssl_required,
                autocommit=True,
            )
            return conn
        except Error as e:
            raise RuntimeError(f"MySQL connection failed: {e}")

    def close(self):
        try:
            if self._conn and self._conn.is_connected():
                self._conn.close()
        except Exception:
            pass

    def _ensure_connection(self):
        if not self._conn or not self._conn.is_connected():
            raise RuntimeError("MySQL connection is not available.")

    def get_node_degree(self, node_id: str) -> int:
        self._ensure_connection()
        try:
            with self._conn.cursor() as cursor:
                query = "SELECT neighbor_counts FROM neighbors WHERE id = %s"
                cursor.execute(query, (node_id,))
                result = cursor.fetchone()
        except Exception as e:
            raise Exception(f"{e}, mysql query failed (get_node_degree)")

        if result and result[0]:
            # neighbor_counts may come back as str, bytes, or dict depending on column type/driver settings
            neighbor_counts = result[0]
            if isinstance(neighbor_counts, (bytes, bytearray)):
                neighbor_counts = neighbor_counts.decode("utf-8")

            if isinstance(neighbor_counts, str):
                degree_by_biolink_type = json.loads(neighbor_counts)
            elif isinstance(neighbor_counts, dict):
                degree_by_biolink_type = neighbor_counts
            else:
                # last resort
                degree_by_biolink_type = json.loads(str(neighbor_counts))

            return int(degree_by_biolink_type.get("biolink:NamedThing", 0))

        return 0

    def get_degrees_by_node(self, curie_ids: List[str], batch_size: int = 10000) -> Dict[str, dict]:
        self._ensure_connection()
        if not curie_ids:
            return {}

        degree_dict: Dict[str, dict] = {}

        try:
            with self._conn.cursor() as cursor:
                for i in range(0, len(curie_ids), batch_size):
                    batch_ids = curie_ids[i : i + batch_size]
                    placeholders = ",".join(["%s"] * len(batch_ids))
                    query = f"""
                        SELECT id, neighbor_counts
                        FROM neighbors
                        WHERE id IN ({placeholders})
                    """
                    cursor.execute(query, tuple(batch_ids))

                    for node_id, neighbor_counts in cursor.fetchall():
                        if isinstance(neighbor_counts, (bytes, bytearray)):
                            neighbor_counts = neighbor_counts.decode("utf-8")

                        if isinstance(neighbor_counts, str):
                            degree_by_biolink_type = json.loads(neighbor_counts)
                        elif isinstance(neighbor_counts, dict):
                            degree_by_biolink_type = neighbor_counts
                        else:
                            degree_by_biolink_type = json.loads(str(neighbor_counts))

                        degree_dict[node_id] = degree_by_biolink_type

        except Exception as e:
            raise Exception(f"{e}, mysql query failed (get_degrees_by_node)")

        for curie in curie_ids:
            degree_dict.setdefault(curie, {})

        return degree_dict

    def get_degree_categories(self, batch_size: int = 10000) -> Set[str]:
        self._ensure_connection()

        degree_category_set: Set[str] = set()
        offset = 0

        try:
            with self._conn.cursor() as cursor:
                while True:
                    query = f"""
                        SELECT neighbor_counts
                        FROM neighbors
                        LIMIT {batch_size} OFFSET {offset}
                    """
                    cursor.execute(query)
                    rows = cursor.fetchall()

                    if not rows:
                        break

                    for (neighbor_counts,) in rows:
                        if isinstance(neighbor_counts, (bytes, bytearray)):
                            neighbor_counts = neighbor_counts.decode("utf-8")

                        if isinstance(neighbor_counts, str):
                            degree_by_biolink_type = json.loads(neighbor_counts)
                        elif isinstance(neighbor_counts, dict):
                            degree_by_biolink_type = neighbor_counts
                        else:
                            degree_by_biolink_type = json.loads(str(neighbor_counts))

                        degree_category_set.update(degree_by_biolink_type.keys())

                    offset += batch_size

        except Exception as e:
            raise Exception(f"{e}, mysql query failed (get_degree_categories)")

        return degree_category_set



from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class MySQLConfig:
    host: str
    user: str
    database: str
    password: Optional[str] = None
    ssl_required: bool = True
    port: int = 3306


def parse_mysql_config(config_str: str) -> MySQLConfig:
    """
    Expected formats:
      mysql:host:user:database
      mysql:host:user:password:database
      mysql:host:port:user:database
      mysql:host:port:user:password:database
    """

    if not config_str:
        raise ValueError("Empty MySQL config string")

    parts = config_str.split(":")

    if parts[0].lower() != "mysql":
        raise ValueError(f"Unsupported config scheme: {parts[0]}")

    # mysql:host:user:db
    if len(parts) == 4:
        _, host, user, database = parts
        return MySQLConfig(
            host=host,
            user=user,
            database=database,
        )

    # mysql:host:user:password:db
    if len(parts) == 5 and parts[2].isdigit() is False:
        _, host, user, password, database = parts
        return MySQLConfig(
            host=host,
            user=user,
            password=password,
            database=database,
        )

    # mysql:host:port:user:db
    if len(parts) == 5 and parts[2].isdigit():
        _, host, port, user, database = parts
        return MySQLConfig(
            host=host,
            port=int(port),
            user=user,
            database=database,
        )

    # mysql:host:port:user:password:db
    if len(parts) == 6:
        _, host, port, user, password, database = parts
        return MySQLConfig(
            host=host,
            port=int(port),
            user=user,
            password=password,
            database=database,
        )

    raise ValueError(
        "Invalid MySQL config format. "
        "Expected mysql:host:user:db "
        "or mysql:host:user:password:db "
        "or mysql:host:port:user:db "
        "or mysql:host:port:user:password:db"
    )